import re
import types
from copy import copy
from django.template import Library, Node, TemplateSyntaxError

register = Library()


def silence_without_field(fn):
    def wrapped(field, attr):
        if not field:
            return ""
        return fn(field, attr)

    return wrapped


def _process_field_attributes(field, attr, process):
    # split attribute name and value from 'attr:value' string
    # params = attr.split(':', 1)
    # attribute = params[0]
    params = re.split(r"(?<!:):(?!:)", attr, maxsplit=1)
    # attribute = params[0]
    attribute = params[0].replace("::", ":")
    value = params[1] if len(params) == 2 else True
    field = copy(field)

    if not hasattr(field, "as_widget"):
        old_tag = field.tag

        def tag(self, wrap_label=False):  # pylint: disable=unused-argument
            attrs = self.data["attrs"]
            process(self.parent_widget, attrs, attribute, value)
            html = old_tag(wrap_label=False)
            self.tag = old_tag
            return html

        field.tag = types.MethodType(tag, field)
        return field

    # decorate field.as_widget method with updated attributes
    old_as_widget = field.as_widget

    def as_widget(self, widget=None, attrs=None, only_initial=False):
        attrs = attrs or {}
        process(widget or self.field.widget, attrs, attribute, value)
        if attribute == "type":  # change the Input type
            self.field.widget.input_type = value
            del attrs["type"]
        html = old_as_widget(widget, attrs, only_initial)
        self.as_widget = old_as_widget
        return html

    field.as_widget = types.MethodType(as_widget, field)
    return field


@register.filter("attr")
@silence_without_field
def set_attr(field, attr):
    def process(widget, attrs, attribute, value):  # pylint: disable=unused-argument
        attrs[attribute] = value

    return _process_field_attributes(field, attr, process)


@register.filter("add_error_attr")
@silence_without_field
def add_error_attr(field, attr):
    if hasattr(field, "errors") and field.errors:
        return set_attr(field, attr)
    return field


@register.filter("append_attr")
@silence_without_field
def append_attr(field, attr):
    def process(widget, attrs, attribute, value):
        if attrs.get(attribute):
            attrs[attribute] += " " + value
        elif widget.attrs.get(attribute):
            attrs[attribute] = widget.attrs[attribute] + " " + value
        else:
            attrs[attribute] = value

    return _process_field_attributes(field, attr, process)


@register.filter("add_class")
@silence_without_field
def add_class(field, css_class):
    return append_attr(field, "class:" + css_class)


@register.filter("add_label_class")
@silence_without_field
def add_label_class(field, css_class):
    return field.label_tag(attrs={"class": css_class})


@register.filter("add_error_class")
@silence_without_field
def add_error_class(field, css_class):
    if hasattr(field, "errors") and field.errors:
        return add_class(field, css_class)
    return field


@register.filter("add_required_class")
@silence_without_field
def add_required_class(field, css_class):
    if hasattr(field.field, "required") and field.field.required:
        return add_class(field, css_class)
    return field


@register.filter("set_data")
@silence_without_field
def set_data(field, data):
    return set_attr(field, "data-" + data)


@register.filter(name="field_type")
def field_type(field):
    """
    Template filter that returns field class name (in lower case).
    E.g. if field is CharField then {{ field|field_type }} will
    return 'charfield'.
    """
    if hasattr(field, "field") and field.field:
        return field.field.__class__.__name__.lower()
    return ""


@register.filter(name="widget_type")
def widget_type(field):
    """
    Template filter that returns field widget class name (in lower case).
    E.g. if field's widget is TextInput then {{ field|widget_type }} will
    return 'textinput'.
    """
    if (
        hasattr(field, "field")
        and hasattr(field.field, "widget")
        and field.field.widget
    ):
        return field.field.widget.__class__.__name__.lower()
    return ""


# ======================== render_field tag ==============================

ATTRIBUTE_RE = re.compile(
    r"""
    (?P<attr>
        [@\w:_\.-]+
    )
    (?P<sign>
        \+?=
    )
    (?P<value>
    ['"]? # start quote
        [^"']*
    ['"]? # end quote
    )
""",
    re.VERBOSE | re.UNICODE,
)


@register.tag
def render_field(parser, token):  # pylint: disable=too-many-locals
    """
    Render a form field using given attribute-value pairs

    Takes form field as first argument and list of attribute-value pairs for
    all other arguments.  Attribute-value pairs should be in the form of
    attribute=value or attribute="a value" for assignment and attribute+=value
    or attribute+="value" for appending.
    """
    error_msg = (
        f"{token.split_contents()[0]!r} tag requires a form field followed by "
        'a list of attributes and values in the form attr="value"'
    )
    try:
        bits = token.split_contents()
        tag_name = bits[0]  # noqa
        form_field = bits[1]
        attr_list = bits[2:]
    except ValueError as exc:
        raise TemplateSyntaxError(error_msg) from exc

    form_field = parser.compile_filter(form_field)

    set_attrs = []
    append_attrs = []
    for pair in attr_list:
        match = ATTRIBUTE_RE.match(pair)
        if not match:
            raise TemplateSyntaxError(error_msg + f": {pair}")
        dct = match.groupdict()
        attr, sign, value = (
            dct["attr"],
            dct["sign"],
            parser.compile_filter(dct["value"]),
        )
        if sign == "=":
            set_attrs.append((attr, value))
        else:
            append_attrs.append((attr, value))

    return FieldAttributeNode(form_field, set_attrs, append_attrs)


class FieldAttributeNode(Node):
    def __init__(self, field, set_attrs, append_attrs):
        self.field = field
        self.set_attrs = set_attrs
        self.append_attrs = append_attrs

    def render(self, context):
        bounded_field = self.field.resolve(context)
        field = getattr(bounded_field, "field", None)
        if getattr(bounded_field, "errors", None) and "WIDGET_ERROR_CLASS" in context:
            bounded_field = append_attr(
                bounded_field, f'class:{context["WIDGET_ERROR_CLASS"]}'
            )
        if field and field.required and "WIDGET_REQUIRED_CLASS" in context:
            bounded_field = append_attr(
                bounded_field, f"class:{context['WIDGET_REQUIRED_CLASS']}"
            )
        for k, v in self.set_attrs:
            if k == "type":
                bounded_field.field.widget.input_type = v.resolve(context)
            else:
                bounded_field = set_attr(bounded_field, f"{k}:{v.resolve(context)}")
        for k, v in self.append_attrs:
            bounded_field = append_attr(bounded_field, f"{k}:{v.resolve(context)}")
        return str(bounded_field)


# ======================== remove_attr tag ==============================


@register.filter("remove_attr")
@silence_without_field
def remove_attr(field, attr):
    if attr in field.field.widget.attrs:
        del field.field.widget.attrs[attr]
    return field



from unittest import TestCase

from .forms import (
    render_field,
    render_choice_field,
    render_field_from_tag,
    render_form,
    MyForm,
)


def assertIn(value, obj):
    assert value in obj, f"{value} not in {obj}"


def assertNotIn(value, obj):
    assert value not in obj, f"{value} in {obj}"


# ===============================
#           Test cases
# ===============================


class SimpleAttrTest(TestCase):
    def test_attr(self):
        res = render_field("simple", "attr", "foo:bar")
        assertIn('type="text"', res)
        assertIn('name="simple"', res)
        assertIn('id="id_simple"', res)
        assertIn('foo="bar"', res)

    def test_attr_chaining(self):
        res = render_field("simple", "attr", "foo:bar", "attr", "bar:baz")
        assertIn('type="text"', res)
        assertIn('name="simple"', res)
        assertIn('id="id_simple"', res)
        assertIn('foo="bar"', res)
        assertIn('bar="baz"', res)

    def test_add_class(self):
        res = render_field("simple", "add_class", "foo")
        assertIn('class="foo"', res)

    def test_add_multiple_classes(self):
        res = render_field("simple", "add_class", "foo bar")
        assertIn('class="foo bar"', res)

    def test_add_class_chaining(self):
        res = render_field("simple", "add_class", "foo", "add_class", "bar")
        assertIn('class="bar foo"', res)

    def test_set_data(self):
        res = render_field("simple", "set_data", "key:value")
        assertIn('data-key="value"', res)

    def test_replace_type(self):
        res = render_field("simple", "attr", "type:date")
        self.assertTrue(res.count("type=") == 1, (res, res.count("type=")))
        assertIn('type="date"', res)

    def test_replace_hidden(self):
        res = render_field("simple", "attr", "type:hidden")
        self.assertTrue(res.count("type=") == 1, (res, res.count("type=")))
        assertIn('type="hidden"', res)


class ErrorsTest(TestCase):
    def _err_form(self):
        form = MyForm({"foo": "bar"})  # some random data
        form.is_valid()  # trigger form validation
        return form

    def test_error_class_no_error(self):
        res = render_field("simple", "add_error_class", "err")
        assertNotIn('class="err"', res)

    def test_error_class_error(self):
        form = self._err_form()
        res = render_field("simple", "add_error_class", "err", form=form)
        assertIn('class="err"', res)

    def test_required_class(self):
        res = render_field("simple", "add_required_class", "is-required")
        assertIn('class="is-required"', res)

    def test_required_class_requiredfield(self):
        form = self._err_form()
        res = render_field("simple", "add_required_class", "is-required", form=form)
        assertIn('class="is-required"', res)
        assertIn("required", res)

    def test_error_attr_no_error(self):
        res = render_field("simple", "add_error_attr", "aria-invalid:true")
        assertNotIn('aria-invalid="true"', res)

    def test_error_attr_error(self):
        form = self._err_form()
        res = render_field("simple", "add_error_attr", "aria-invalid:true", form=form)
        assertIn('aria-invalid="true"', res)


class SilenceTest(TestCase):
    def test_silence_without_field(self):
        res = render_field("nothing", "attr", "foo:bar")
        self.assertEqual(res, "")
        res = render_field("nothing", "add_class", "some")
        self.assertEqual(res, "")
        res = render_field("nothing", "remove_attr", "some")
        self.assertEqual(res, "")


class CustomizedWidgetTest(TestCase):
    def test_attr(self):
        res = render_field("with_attrs", "attr", "foo:bar")
        assertIn('foo="bar"', res)
        assertNotIn('foo="baz"', res)
        assertIn('egg="spam"', res)

    def test_attr_chaining(self):
        res = render_field("with_attrs", "attr", "foo:bar", "attr", "bar:baz")
        assertIn('foo="bar"', res)
        assertNotIn('foo="baz"', res)
        assertIn('egg="spam"', res)
        assertIn('bar="baz"', res)

    def test_attr_class(self):
        res = render_field("with_cls", "attr", "foo:bar")
        assertIn('foo="bar"', res)
        assertIn('class="class0"', res)

    def test_default_attr(self):
        res = render_field("with_cls", "attr", "type:search")
        assertIn('class="class0"', res)
        assertIn('type="search"', res)

    def test_add_class(self):
        res = render_field("with_cls", "add_class", "class1")
        assertIn("class0", res)
        assertIn("class1", res)

    def test_add_class_chaining(self):
        res = render_field("with_cls", "add_class", "class1", "add_class", "class2")
        assertIn("class0", res)
        assertIn("class1", res)
        assertIn("class2", res)

    def test_remove_attr(self):
        res = render_field("with_attrs", "remove_attr", "foo")
        assertNotIn("foo", res)


class FieldReuseTest(TestCase):
    def test_field_double_rendering_simple(self):
        res = render_form(
            '{{ form.simple }}{{ form.simple|attr:"foo:bar" }}{{ form.simple }}'
        )
        self.assertEqual(res.count("bar"), 1)

    def test_field_double_rendering_simple_css(self):
        res = render_form(
            '{{ form.simple }}{{ form.simple|add_class:"bar" }}{{ form.simple|add_class:"baz" }}'
        )
        self.assertEqual(res.count("baz"), 1)
        self.assertEqual(res.count("bar"), 1)

    def test_field_double_rendering_attrs(self):
        res = render_form(
            '{{ form.with_cls }}{{ form.with_cls|add_class:"bar" }}{{ form.with_cls }}'
        )
        self.assertEqual(res.count("class0"), 3)
        self.assertEqual(res.count("bar"), 1)


class SimpleRenderFieldTagTest(TestCase):
    def test_attr(self):
        res = render_field_from_tag("simple", 'foo="bar"')
        assertIn('type="text"', res)
        assertIn('name="simple"', res)
        assertIn('id="id_simple"', res)
        assertIn('foo="bar"', res)

    def test_multiple_attrs(self):
        res = render_field_from_tag("simple", 'foo="bar"', 'bar="baz"')
        assertIn('type="text"', res)
        assertIn('name="simple"', res)
        assertIn('id="id_simple"', res)
        assertIn('foo="bar"', res)
        assertIn('bar="baz"', res)


class RenderFieldTagSilenceTest(TestCase):
    def test_silence_without_field(self):
        res = render_field_from_tag("nothing", 'foo="bar"')
        self.assertEqual(res, "")
        res = render_field_from_tag("nothing", 'class+="some"')
        self.assertEqual(res, "")


class RenderFieldTagCustomizedWidgetTest(TestCase):
    def test_attr(self):
        res = render_field_from_tag("with_attrs", 'foo="bar"')
        assertIn('foo="bar"', res)
        assertNotIn('foo="baz"', res)
        assertIn('egg="spam"', res)

    def test_multiple_attrs(self):
        res = render_field_from_tag("with_attrs", 'foo="bar"', 'bar="baz"')
        assertIn('foo="bar"', res)
        assertNotIn('foo="baz"', res)
        assertIn('egg="spam"', res)
        assertIn('bar="baz"', res)

    def test_attr_class(self):
        res = render_field_from_tag("with_cls", 'foo="bar"')
        assertIn('foo="bar"', res)
        assertIn('class="class0"', res)

    def test_default_attr(self):
        res = render_field_from_tag("with_cls", 'type="search"')
        assertIn('class="class0"', res)
        assertIn('type="search"', res)

    def test_append_attr(self):
        res = render_field_from_tag("with_cls", 'class+="class1"')
        assertIn("class0", res)
        assertIn("class1", res)

    def test_duplicate_append_attr(self):
        res = render_field_from_tag("with_cls", 'class+="class1"', 'class+="class2"')
        assertIn("class0", res)
        assertIn("class1", res)
        assertIn("class2", res)

    def test_hyphenated_attributes(self):
        res = render_field_from_tag("with_cls", 'data-foo="bar"')
        assertIn('data-foo="bar"', res)
        assertIn('class="class0"', res)

    def test_alpinejs_event_modifier(self):
        res = render_field_from_tag(
            "simple", '@click.away="open=false"', 'x-on::click.away="open=false"'
        )
        assertIn('@click.away="open=false"', res)
        assertIn('x-on:click.away="open=false"', res)


class RenderFieldWidgetClassesTest(TestCase):
    def test_use_widget_required_class(self):
        res = render_form(
            "{% render_field form.simple %}", WIDGET_REQUIRED_CLASS="required_class"
        )
        assertIn('class="required_class"', res)

    def test_use_widget_error_class(self):
        res = render_form(
            "{% render_field form.simple %}",
            form=MyForm({}),
            WIDGET_ERROR_CLASS="error_class",
        )
        assertIn('class="error_class"', res)

    def test_use_widget_error_class_with_other_classes(self):
        res = render_form(
            '{% render_field form.simple class="blue" %}',
            form=MyForm({}),
            WIDGET_ERROR_CLASS="error_class",
        )
        assertIn('class="blue error_class"', res)

    def test_use_widget_required_class_with_other_classes(self):
        res = render_form(
            '{% render_field form.simple class="blue" %}',
            form=MyForm({}),
            WIDGET_REQUIRED_CLASS="required_class",
        )
        assertIn('class="blue required_class"', res)


class RenderFieldTagFieldReuseTest(TestCase):
    def test_field_double_rendering_simple(self):
        res = render_form(
            '{{ form.simple }}{% render_field form.simple foo="bar" %}{% render_field form.simple %}'
        )
        self.assertEqual(res.count("bar"), 1)

    def test_field_double_rendering_simple_css(self):
        res = render_form(
            '{% render_field form.simple %}{% render_field form.simple class+="bar" %}{% render_field form.simple class+="baz" %}'
        )
        self.assertEqual(res.count("baz"), 1)
        self.assertEqual(res.count("bar"), 1)

    def test_field_double_rendering_attrs(self):
        res = render_form(
            '{% render_field form.with_cls %}{% render_field form.with_cls class+="bar" %}{% render_field form.with_cls %}'
        )
        self.assertEqual(res.count("class0"), 3)
        self.assertEqual(res.count("bar"), 1)

    def test_field_double_rendering_id(self):
        res = render_form(
            "{{ form.simple }}"
            '{% render_field form.simple id="id_1" %}'
            '{% render_field form.simple id="id_2" %}'
        )
        self.assertEqual(res.count("id_1"), 1)
        self.assertEqual(res.count("id_2"), 1)

    def test_field_double_rendering_id_name(self):
        res = render_form(
            "{{ form.simple }}"
            '{% render_field form.simple id="id_1" name="n_1" %}'
            '{% render_field form.simple id="id_2" name="n_2" %}'
        )
        self.assertEqual(res.count("id_1"), 1)
        self.assertEqual(res.count("id_2"), 1)
        self.assertEqual(res.count("n_1"), 1)
        self.assertEqual(res.count("n_2"), 1)

    def test_field_double_rendering_id_class(self):
        res = render_form(
            "{{ form.simple }}"
            '{% render_field form.simple id="id_1" class="c_1" %}'
            '{% render_field form.simple id="id_2" class="c_2" %}'
        )
        self.assertEqual(res.count("id_1"), 1)
        self.assertEqual(res.count("id_2"), 1)
        self.assertEqual(res.count("c_1"), 1)
        self.assertEqual(res.count("c_2"), 1)

    def test_field_double_rendering_name_class(self):
        res = render_form(
            "{{ form.simple }}"
            '{% render_field form.simple name="n_1" class="c_1" %}'
            '{% render_field form.simple name="n_2" class="c_2" %}'
        )
        self.assertEqual(res.count("n_1"), 1)
        self.assertEqual(res.count("n_2"), 1)
        self.assertEqual(res.count("c_1"), 1)
        self.assertEqual(res.count("c_2"), 1)

    def test_field_double_rendering_simple_again(self):
        res = render_form('{% render_field form.simple foo="bar" v-model="username" %}')
        self.assertEqual(res.count('v-model="username"'), 1)


class RenderFieldTagUseTemplateVariableTest(TestCase):
    def test_use_template_variable_in_parameters(self):
        res = render_form(
            '{% render_field form.with_attrs egg+="pahaz" placeholder=form.with_attrs.label %}'
        )
        assertIn('egg="spam pahaz"', res)
        assertIn('placeholder="With attrs"', res)


class RenderFieldFilter_field_type_widget_type_Test(TestCase):
    def test_field_type_widget_type_rendering_simple(self):
        res = render_form(
            '<div class="{{ form.simple|field_type }} {{ form.simple|widget_type }} {{ form.simple.html_name }}">{{ form.simple }}</div>'
        )
        assertIn('class="charfield textinput simple"', res)


class RenderFieldTagNonValueAttribute(TestCase):
    def test_field_non_value(self):
        res = render_form('{{ form.simple|attr:"foo" }}')
        assertIn("foo", res)
        assertNotIn("foo=", res)

    def test_field_empty_value(self):
        res = render_form('{{ form.simple|attr:"foo:" }}')
        assertIn('foo=""', res)

    def test_field_other_value(self):
        res = render_form('{{ form.simple|attr:"foo:bar" }}')
        assertIn('foo="bar"', res)

    def test_field_double_colon(self):
        res = render_form('{{ form.simple|attr:"v-bind::class:value" }}')
        assertIn('v-bind:class="value"', res)

    def test_field_double_colon_morethanone(self):
        res = render_form('{{ form.simple|attr:"v-bind::class:{active:True}" }}')
        assertIn('v-bind:class="{active:True}"', res)

    def test_field_arroba(self):
        res = render_form('{{ form.simple|attr:"@click:onClick" }}')
        assertIn('@click="onClick"', res)

    def test_field_arroba_dot(self):
        res = render_form('{{ form.simple|attr:"@click.prevent:onClick" }}')
        assertIn('@click.prevent="onClick"', res)

    def test_field_double_colon_missing(self):
        res = render_form('{{ form.simple|attr:"::class:{active:True}" }}')
        assertIn(':class="{active:True}"', res)


class SelectFieldTest(TestCase):
    def test_parent_field(self):
        res = render_field("choice", "attr", "foo:bar")
        assertIn("select", res)
        assertIn('name="choice"', res)
        assertIn('id="id_choice"', res)
        assertIn('foo="bar"', res)
        assertIn('<option value="1">one</option>', res)
        assertIn('<option value="2">two</option>', res)

    def test_rendering_id_class(self):
        res = render_form(
            '{% render_field form.choice id="id_1" class="c_1" %}'
            '{% render_field form.choice id="id_2" class="c_2" %}'
        )
        self.assertEqual(res.count("id_1"), 1)
        self.assertEqual(res.count("id_2"), 1)
        self.assertEqual(res.count("c_1"), 1)
        self.assertEqual(res.count("c_2"), 1)


class RadioFieldTest(TestCase):
    def test_first_choice(self):
        res = render_choice_field("radio", 0, "attr", "foo:bar")
        assertIn('type="radio"', res)
        assertIn('name="radio"', res)
        assertIn('value="option1"', res)
        assertIn('id="id_radio_0"', res)
        assertIn('foo="bar"', res)

    def test_second_choice(self):
        res = render_choice_field("radio", 1, "attr", "foo:bar")
        assertIn('type="radio"', res)
        assertIn('name="radio"', res)
        assertIn('value="option2"', res)
        assertIn('id="id_radio_1"', res)
        assertIn('foo="bar"', res)

    def test_rendering_id_class(self):
        res = render_form(
            '{% render_field form.radio.0 id="id_1" class="c_1" %}'
            '{% render_field form.radio.1 id="id_2" class="c_2" %}'
        )
        self.assertEqual(res.count("id_1"), 1)
        self.assertEqual(res.count("id_2"), 1)
        self.assertEqual(res.count("c_1"), 1)
        self.assertEqual(res.count("c_2"), 1)



from importlib.metadata import version, PackageNotFoundError

try:
    __version__ = version("django-widget-tweaks")
except PackageNotFoundError:
    # package is not installed
    __version__ = None
